#!/usr/bin/env python3
# encoding: utf-8
#
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2024 Henrik tom Wörden
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#

"""
A data model class for the semantic data that shall be created by synchronization of the crawler.
"""

from typing import Any, Dict, List, Optional, Union
from uuid import uuid4 as uuid

import linkahead as db
from linkahead.apiutils import (EntityMergeConflictError, compare_entities,
                                merge_entities)
from linkahead.cached import cache_clear, cached_get_entity_by
from linkahead.exceptions import EmptyUniqueQueryError

from .identifiable_adapters import IdentifiableAdapter


class SemanticEntity():
    """ represents the information related to an Entity as it shall be created in LinkAhead

    Parents and Properties are given by fragments: db.Record objects that may have been created by
    the scanner.
    """

    def __init__(self, entity: db.Entity, registered_identifiable):
        self.fragments = [entity]
        self.id = entity.id
        self.path = entity.path
        self.identifiable = None
        self.registered_identifiable = registered_identifiable
        self.uuid = uuid()

    def identify_with(self, remote_entity):
        """ the given remote_entity is considered to be the target entity.

        ID and path are copied to this objects attributes and to its fragments
        """

        self.id = remote_entity.id
        self.path = remote_entity.path

        for f in self.fragments:
            # side effect
            f.id = remote_entity.id
            f.path = remote_entity.path
            # TODO check the following copying of _size and _checksum
            # Copy over checksum and size too if it is a file
            f._size = remote_entity._size
            f._checksum = remote_entity._checksum

    def include(self, source):
        self.fragments.extend(source.fragments)

        if source.id is not None:
            if self.id is None:
                self.id = source.id
            else:
                assert self.id == source.id

        if source.path is not None:
            if self.path is None:
                self.path = source.path
            else:
                assert self.path == source.path


class SemanticTarget():
    """ models the target structure of Entities as it shall be created by the Crawler

    The target entities are composed using the information of the entity fragments (db.Entity
    objects) of SemanticEntities. This is information like name, parents and properties and,
    importantly, references. Those references are typically given by a Python reference to some
    other db.Entity object. These references are scanned initially and used to create multiple
    reference maps that track how SemanticEntities reference each other. These maps are kept up to
    date when SemanticEntities are merged because they are identified with each other. When
    creating the final list of db.Entity objects (``create_record_lists``) the Python references
    are updated according to the reference map.

    This model should only be manipulated via three functions:
    - make_identifiable: adds an identifiable to a SemanticEntity what possibly allows to merge it
                         with another SemanticEntity
    - merge_with_equivalent: check whether there is an equivalent SemanticEntity and merge into
                             that if one is found
    - check_remote_server: uses the identifiable to check the remote server and add ID and path if
                           an object was found
    - set_existing: declares that a SemanticEntity is existing on the remote server
    - set_missing: declares that a SemanticEntity is NOT existing on the remote server
    """

    def __init__(self, entities: List[db.Entity], identifiableAdapter):
        self.identifiableAdapter = identifiableAdapter
        self._id_look_up: Dict[int, SemanticEntity] = {}
        self._path_look_up: Dict[str, SemanticEntity] = {}
        self._identifiable_look_up: Dict[str, SemanticEntity] = {}
        self._missing: Dict[int, SemanticEntity] = {}
        self._existing: Dict[int, SemanticEntity] = {}
        # entities that are missing get negative IDs to allow identifiable creation
        self._remote_missing_counter = -1

        # create initial set of SemanticEntities from provided Entity list
        self.se: List[SemanticEntity] = []  # list of all SemanticEntities
        # TODO do we only need this for creating the initial reference map? Remove it?
        self.se_lookup: Dict[str, SemanticEntity] = {}  # lookup: UUID -> SemanticEntity
        entities = self._create_flat_list(entities)
        self._sanity_check(entities)
        for el in entities:
            self.se.append(SemanticEntity(
                el,
                self.identifiableAdapter.get_registered_identifiable(el)))
            self.se_lookup[id(el)] = self.se[-1]
        self.unchecked = list(self.se)  # list all SemanticEntities that have not yet been checked

        # initialize reference mappings
        (
            self.forward_references,
            self.backward_references,
            self.forward_id_references,
            self.backward_id_references,
            self.forward_id_referenced_by,
            self.backward_id_referenced_by,
        ) = self._create_reference_mapping(self.se, self.se_lookup)

        self._mark_entities_with_path_or_id()

    def make_identifiable(self, se: SemanticEntity):
        """ creates an identifiable for the given SemanticEntity and possibly merges it into an
        equivalent SemanticEntity
        """
        if se.identifiable is not None:
            raise RuntimeError("Already has identifiable")
        se.identifiable = self.identifiableAdapter.get_identifiable(
            se, self.backward_id_referenced_by[se.uuid])

    def merge_with_equivalent(self, se: SemanticEntity):
        equivalent_se = self.get_checked_equivalent(se)
        if equivalent_se is None:
            return False
        else:
            self._merge_into(se, equivalent_se)
            return True

    def check_remote_server(self, se: SemanticEntity):
        identified_record = (
            self.identifiableAdapter.retrieve_identified_record_for_identifiable(
                se.identifiable))
        if identified_record is not None:
            se.identify_with(identified_record)

    def set_missing(self, se: SemanticEntity):
        """ add the given SemanticEntity to the list of missing entities

        This removes the SemanticEntity from the unchecked list and implies that the entity does
        NOT exist on the remote server.
        """
        assert se.id is None
        if se.path is None and se.identifiable is None:
            raise RuntimeError("no identifying information")
        se.id = self._remote_missing_counter
        for f in se.fragments:
            f.id = self._remote_missing_counter
        self._remote_missing_counter -= 1
        self._add_any(se, self._missing)
        self.unchecked.remove(se)

    def set_existing(self, se: SemanticEntity):
        """ add the given SemanticEntity to the list of existing entities

        This removes the SemanticEntity from the unchecked list and implies that the entity exists
        on the remote server.
        """
        assert se.id is not None
        assert se.id > 0
        self._add_any(se, self._existing)
        self.unchecked.remove(se)

    def get_checked_equivalent(self, entity: SemanticEntity) -> Optional[SemanticEntity]:
        """
        Return an equivalent SemanticEntity from the list of missing or existing entities.

        Equivalent means that ID, path or identifiable are the same.
        """
        # TODO shall we also provide a variant that returns equivalent objects that are unchecked?
        if entity.id is not None and entity.id in self._id_look_up:
            return self._id_look_up[entity.id]
        if entity.path is not None and entity.path in self._path_look_up:
            return self._path_look_up[entity.path]
        if (entity.identifiable is not None and entity.identifiable.get_representation() in
                self._identifiable_look_up):
            return self._identifiable_look_up[entity.identifiable.get_representation()]

    def get_equivalent_existing(self, se):
        """ Check whether this Record exists on the remote server

        Returns: The stored Record
        """
        treated = self.get_any(se)
        if id(treated) in self._existing:
            return rec
        else:
            return None

    def get_equivalent_missing(self, se):
        """ Check whether this Record is missing on the remote server

        Returns: The stored Record
        """
        treated = self.get_any(record, identifiable)
        if id(treated) in self._missing:
            return rec
        else:
            return None

    def combine_fragments(self):
        for se in self.se:
            if len(se.fragments) < 2:
                continue
            for ent in se.fragments[1:]:
                merge_entities(se.fragments[0], ent, merge_id_with_resolved_entity=True)

    def create_record_lists(self):
        for se in self.se:
            for f in se.fragments:
                f.id = se.id
                f.path = se.path
        self._update_reference_values()
        self.combine_fragments()
        # TODO assure that there is only one fragment each

        missing = [el.fragments[0] for el in self._missing.values()]
        # remove negative IDs
        for el in missing:
            if el.id is None:
                raise RuntimeError("This should not happen")  # TODO remove
            if el.id >= 0:
                raise RuntimeError("This should not happen")  # TODO remove
            el.id = None

        return (missing, [el.fragments[0] for el in self._existing.values()])

    def _update_reference_values(self):
        for se in self.se:
            for f in se.fragments:
                for p in f.properties:
                    if isinstance(p.value, list):
                        for index, val in enumerate(p.value):
                            if id(val) in self.se_lookup:
                                p.value[index] = self.se_lookup[id(val)].fragments[0]
                    else:
                        if id(p.value) in self.se_lookup:
                            p.value = self.se_lookup[id(p.value)].fragments[0]

    def identity_relies_on_unchecked_entity(self, se: SemanticEntity):
        """
        If a record for which it could not yet be verified whether it exists in LA or not is part
        of the identifying properties, this returns True, otherwise False
        """

        return any([id(ent) not in self._missing and id(ent) not in self._existing
                    for ent in self.forward_id_references[se.uuid]]
                   + [id(ent) not in self._missing and id(ent) not in self._existing
                      for ent in self.backward_id_referenced_by[se.uuid]])

    def identity_relies_on_missing_entity(self, se: SemanticEntity):
        """
        returns False if any value in the properties attribute is a db.Entity object that
        is contained in the `remote_missing_cache`. If ident has such an object in
        properties, it means that it references another Entity, where we checked
        whether it exists remotely and it was not found.
        """
        return any([id(ent) in self._missing for ent in self.forward_id_references[se.uuid]]
                   + [id(ent) in self._missing for ent in self.backward_id_referenced_by[se.uuid]])

    @ staticmethod
    def _sanity_check(entities: List[db.Entity]):
        for ent in entities:
            if ent.role == "Record" and len(ent.parents) == 0:
                raise RuntimeError(f"Records must have a parent.\n{ent}")

    @ staticmethod
    def _create_flat_list(ent_list: List[db.Entity], flat: Optional[List[db.Entity]] = None):
        """
        Recursively adds entities and all their properties contained in ent_list to
        the output list flat.

        TODO: This function will be moved to pylib as it is also needed by the
              high level API.
        """
        # Note: A set would be useful here, but we do not want a random order.
        if flat is None:
            flat = list()
        for el in ent_list:
            if el not in flat:
                flat.append(el)
        for ent in ent_list:
            for p in ent.properties:
                # For lists append each element that is of type Entity to flat:
                if isinstance(p.value, list):
                    for el in p.value:
                        if isinstance(el, db.Entity):
                            if el not in flat:
                                flat.append(el)
                                SemanticTarget._create_flat_list([el], flat)
                elif isinstance(p.value, db.Entity):
                    if p.value not in flat:
                        flat.append(p.value)
                        SemanticTarget._create_flat_list([p.value], flat)
        return flat

    @ staticmethod
    def _treat_merge_error_of(newrecord, record):
        """
        The parameters are two entities that cannot be merged with the merge_entities function.

        # This function checks for two obvious cases where no merge will ever be possible:
        # 1. Two Entities with differing IDs
        # 2. Two non-Entity values which differ

        It creates a more informative logger message and raises an Exception in those cases.
        """
        for this_p in newrecord.properties:
            that_p = record.get_property(this_p.name)

            if that_p is None:
                logger.debug(f"Property {this_p.name} does not exist in the second entity. Note that "
                             "this should not be the reason for the merge conflict.")
                continue

            if (isinstance(this_p.value, db.Entity)
                    and isinstance(that_p.value, db.Entity)):
                if this_p.value.id is not None and that_p.value.id is not None:
                    if this_p.value.id != that_p.value.id:
                        logger.error("The Crawler is trying to merge two entities "
                                     "because they should be the same object (same"
                                     " identifiables), but they reference "
                                     "different Entities with the same Property."
                                     f"Problematic Property: {this_p.name}\n"
                                     f"Referenced Entities: {this_p.value.id} and "
                                     f"{that_p.value.id}\n"
                                     f"{record}\n{newrecord}")
                        raise RuntimeError("Cannot merge Entities")
            elif (not isinstance(this_p.value, db.Entity)
                  and not isinstance(that_p.value, db.Entity)):
                if ((this_p.value != that_p.value)
                    # TODO can we also compare lists?
                    and not isinstance(this_p.value, list)
                        and not isinstance(that_p.value, list)):
                    logger.error(
                        "The Crawler is trying to merge two entities because they should be the same "
                        "object (same identifiables), but they have different values for the same "
                        "Property.\n"
                        f"Problematic Property: {this_p.name}\n"
                        f"Values: {this_p.value} and {that_p.value}\n"
                        f"{record}\n{newrecord}")
                    raise RuntimeError("Cannot merge Entities")

    @ staticmethod
    def _create_reference_mapping(flat: List[SemanticEntity], se_lookup):
        """
        TODO update docstring
        Create a dictionary of dictionaries of the form:
        dict[int, dict[str, list[Union[int,None]]]]

        - The integer index is the Python id of the value object.
        - The string is the name of the first parent of the referencing object.

        Each value objects is taken from the values of all properties from the list flat.

        So the returned mapping maps ids of entities to the ids of objects which are referring
        to them.
        """
        # TODO we need to treat children of RecordTypes somehow.
        forward_references: Dict[str, set[SemanticEntity]] = {}
        backward_references: Dict[str, set[SemanticEntity]] = {}
        forward_id_references: Dict[str, set[SemanticEntity]] = {}
        backward_id_references: Dict[str, set[SemanticEntity]] = {}
        forward_id_referenced_by: Dict[str, set[SemanticEntity]] = {}
        backward_id_referenced_by: Dict[str, set[SemanticEntity]] = {}

        # initialize with empty lists/dict
        for se in flat:
            for ent in se.fragments:
                forward_references[se.uuid] = set()
                backward_references[se.uuid] = set()
                forward_id_references[se.uuid] = set()
                backward_id_references[se.uuid] = set()
                forward_id_referenced_by[se.uuid] = set()
                backward_id_referenced_by[se.uuid] = set()
        for se in flat:
            for ent in se.fragments:
                for p in ent.properties:
                    val = p.value
                    if not isinstance(val, list):
                        val = [val]
                    for v in val:
                        if isinstance(v, db.Entity):
                            vse = se_lookup[id(v)]
                            forward_references[se.uuid].add(vse)
                            backward_references[vse.uuid].add(se)
                            if len([el.name for el in se.registered_identifiable.properties if
                                    el.name == p.name]) > 0:
                                forward_id_references[se.uuid].add(vse)
                                backward_id_references[vse.uuid].add(se)
                            if (vse.registered_identifiable is not None and
                                    IdentifiableAdapter.referencing_entity_has_appropriate_type(
                                    ent.parents, vse.registered_identifiable)):
                                forward_id_referenced_by[se.uuid].add(vse)
                                backward_id_referenced_by[vse.uuid].add(se)

        return (forward_references, backward_references, forward_id_references,
                backward_id_references, forward_id_referenced_by, backward_id_referenced_by,
                )

    def _mark_entities_with_path_or_id(self):
        """ A path or an ID is sufficiently identifying. Thus, those entities can be marked as
        checked """
        for semantic_entity in list(self.se[::-1]):
            assert len(semantic_entity.fragments) == 1
            entity = semantic_entity.fragments[0]
            if entity.id is None and entity.path is None:
                continue
            if entity.path is not None:
                try:
                    existing = cached_get_entity_by(path=entity.path)
                except EmptyUniqueQueryError:
                    existing = None
                if existing is not None:
                    semantic_entity.identify_with(existing)

            # at this point, semantic_entity has an ID if it is existing
            treated_before = self.get_checked_equivalent(semantic_entity)
            if treated_before is None:
                if semantic_entity.id is None or semantic_entity.id < 0:
                    self.set_missing(semantic_entity)
                else:
                    self.set_existing(semantic_entity)
            else:
                self._merge_into(semantic_entity, treated_before)

    def detect_circular_dependency(self):
        """
        Detects whether there are circular references in the given entity list and returns a list
        where the entities are ordered according to the chain of references (and only the entities
        contained in the circle are included. Returns None if no circular dependency is found.

        TODO: for the sake of detecting problems for split_into_inserts_and_updates we should only
        consider references that are identifying properties.
        """
        circle = [self.unchecked[0]]
        closed = False
        while not closed:
            added_to_circle = False
            for referenced in self.forward_references[circle[-1].uuid]:
                if referenced in self.unchecked:
                    if referenced in circle:
                        closed = True
                    circle.append(pval)
                    added_to_circle = True
            if not added_to_circle:
                return None
        return circle

    def _add_any(self, entity: SemanticEntity, lookup):
        if entity.id is not None:
            self._id_look_up[entity.id] = entity
        if entity.path is not None:
            self._path_look_up[entity.path] = entity
        if entity.identifiable is not None:
            self._identifiable_look_up[entity.identifiable.get_representation()] = entity
        lookup[id(entity)] = entity

    def _merge_into(self, source: SemanticEntity, target: SemanticEntity):
        """ tries to merge record into newrecord

        If it fails, record is added to the try_to_merge_later list.
        In any case, references are bent to the newrecord object.

        """
        for f in source.fragments:
            self.se_lookup[id(f)] = target
            f.id = target.id
            f.path = target.path
        target.include(source)
        if target.identifiable is None and not self.identity_relies_on_unchecked_entity(target):
            try:
                self.make_identifiable(target)
                if target not in self.unchecked:
                    self._identifiable_look_up[target.identifiable.get_representation()] = target
            except Exception as es:
                print(es)
                pass

        # update reference mappings
        for se in self.forward_references.pop(source.uuid):
            self.forward_references[target.uuid].add(se)
            self.backward_references[se.uuid].remove(source)
            self.backward_references[se.uuid].add(target)
        for se in self.backward_references.pop(source.uuid):
            self.backward_references[target.uuid].add(se)
            self.forward_references[se.uuid].remove(source)
            self.forward_references[se.uuid].add(target)

        for se in self.forward_id_references.pop(source.uuid):
            self.forward_id_references[target.uuid].add(se)
            self.backward_id_references[se.uuid].remove(source)
            self.backward_id_references[se.uuid].add(target)
        for se in self.backward_id_references.pop(source.uuid):
            self.backward_id_references[target.uuid].add(se)
            self.forward_id_references[se.uuid].remove(source)
            self.forward_id_references[se.uuid].add(target)

        for se in self.forward_id_referenced_by.pop(source.uuid):
            self.forward_id_referenced_by[target.uuid].add(se)
            self.backward_id_referenced_by[se.uuid].remove(source)
            self.backward_id_referenced_by[se.uuid].add(target)
        for se in self.backward_id_referenced_by.pop(source.uuid):
            self.backward_id_referenced_by[target.uuid].add(se)
            self.forward_id_referenced_by[se.uuid].remove(source)
            self.forward_id_referenced_by[se.uuid].add(target)

        # remove unneeded SemanticEntity
        self.se.remove(source)
        if source in self.unchecked:
            self.unchecked.remove(source)
        assert id(source) not in self._missing
        assert id(source) not in self._existing
