#!/usr/bin/env python3
# encoding: utf-8
#
# This file is a part of the LinkAhead Project.
#
# Copyright (C) 2024 Indiscale GmbH <info@indiscale.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program. If not, see <https://www.gnu.org/licenses/>.
#
from functools import partial
from unittest.mock import MagicMock, Mock, patch

import linkahead as db
from caoscrawler.identifiable_adapters import CaosDBIdentifiableAdapter
from caoscrawler.semantic_target import SemanticEntity, SemanticTarget

from test_crawler import basic_retrieve_by_name_mock_up, mock_get_entity_by


def test_create_flat_list():
    a = db.Record()
    b = db.Record()
    a.add_property(name="a", value=a)
    a.add_property(name="b", value=b)
    flat = SemanticTarget._create_flat_list([a])
    assert len(flat) == 2
    assert a in flat
    assert b in flat
    c = db.Record()
    c.add_property(name="a", value=a)
    # This would caus recursion if it is not dealt with properly.
    a.add_property(name="c", value=c)
    flat = SemanticTarget._create_flat_list([c])
    assert len(flat) == 3
    assert a in flat
    assert b in flat
    assert c in flat


def test_create_reference_mapping():
    a = db.Record().add_parent("A")
    b = db.Record(id=132).add_parent("B").add_property('a', a)
    ses = [SemanticEntity(a, db.RecordType().add_property("is_referenced_by", ["B"])),
           SemanticEntity(b, db.RecordType().add_property("a"))]

    (forward_references, backward_references, forward_id_references,
     backward_id_references, forward_id_referenced_by,
     backward_id_referenced_by) = SemanticTarget._create_reference_mapping(
        ses,
        {id(a): ses[0], id(b): ses[1]}
    )
    # test initialization
    assert ses[0].uuid in forward_references
    assert ses[1].uuid in forward_references
    assert ses[0].uuid in backward_references
    assert ses[1].uuid in backward_references
    assert ses[0].uuid in forward_id_references
    assert ses[1].uuid in forward_id_references
    assert ses[0].uuid in backward_id_references
    assert ses[1].uuid in backward_id_references
    assert ses[0].uuid in forward_id_referenced_by
    assert ses[1].uuid in forward_id_referenced_by
    assert ses[0].uuid in backward_id_referenced_by
    assert ses[1].uuid in backward_id_referenced_by

    # a has no ref
    assert len(forward_references[ses[0].uuid]) == 0
    assert backward_references[ses[0].uuid] == set([ses[1]])
    # b does
    assert forward_references[ses[1].uuid] == set([ses[0]])
    assert backward_references[ses[1].uuid] == set()
    # a has no identifying reference
    assert forward_id_references[ses[0].uuid] == set()
    assert backward_references[ses[0].uuid] == set([ses[1]])
    # b has an identifying reference
    assert forward_id_references[ses[1].uuid] == set([ses[0]])
    assert backward_references[ses[1].uuid] == set()
    # a has an identifying back reference
    assert forward_id_referenced_by[ses[0].uuid] == set()
    assert backward_id_referenced_by[ses[0].uuid] == set([ses[1]])
    # b does not
    assert forward_id_referenced_by[ses[1].uuid] == set([ses[0]])
    assert backward_id_referenced_by[ses[1].uuid] == set()


def test_SemanticTarget():
    a = db.Record().add_parent("A")
    ident_a = db.RecordType().add_parent("A").add_property("prop_ident")
    ident_adapter = CaosDBIdentifiableAdapter()
    ident_adapter.register_identifiable("A", ident_a)
    st = SemanticTarget([a], ident_adapter)


def test_merge_into():
    # simplest case: a -> c
    #                b
    #                (a reference c; b does not reference anything; a & b have the same target record)
    c = db.Record(name='c').add_parent("RT2")
    a = db.Record().add_parent("RT1").add_property('RT2', c)
    b = db.Record().add_parent("RT1")

    # We use the reference as identifying reference in both directions. Thus the map is the same
    # for all three categories: references, id_references and id_referenced_by
    ident_a = db.RecordType().add_parent("RT1").add_property("RT2")
    ident_b = db.RecordType().add_parent("RT2").add_property("is_referenced_by", ["RT1"])
    ident_adapter = CaosDBIdentifiableAdapter()
    ident_adapter.register_identifiable("RT1", ident_a)
    ident_adapter.register_identifiable("RT2", ident_b)

    st = SemanticTarget([a, b], ident_adapter)
    se_a = st.se_lookup[id(a)]
    se_b = st.se_lookup[id(b)]
    se_c = st.se_lookup[id(c)]

    # CHECK REFERENCE MAP:
    # c is referenced by a
    assert len(st.forward_references[se_a.uuid]) == 1
    se_c in st.forward_references[se_a.uuid]
    assert len(st.forward_references[se_b.uuid]) == 0
    assert len(st.forward_references[se_c.uuid]) == 0
    assert len(st.backward_references[se_a.uuid]) == 0
    assert len(st.backward_references[se_b.uuid]) == 0
    assert len(st.backward_references[se_c.uuid]) == 1
    se_a in st.backward_references[se_c.uuid]

    assert len(st.forward_id_references[se_a.uuid]) == 1
    se_c in st.forward_id_references[se_a.uuid]
    assert len(st.forward_id_references[se_b.uuid]) == 0
    assert len(st.forward_id_references[se_c.uuid]) == 0
    assert len(st.backward_id_references[se_a.uuid]) == 0
    assert len(st.backward_id_references[se_b.uuid]) == 0
    assert len(st.backward_id_references[se_c.uuid]) == 1
    se_a in st.backward_id_references[se_c.uuid]

    assert len(st.forward_id_referenced_by[se_a.uuid]) == 1
    se_c in st.forward_id_referenced_by[se_a.uuid]
    assert len(st.forward_id_referenced_by[se_b.uuid]) == 0
    assert len(st.forward_id_referenced_by[se_c.uuid]) == 0
    assert len(st.backward_id_referenced_by[se_a.uuid]) == 0
    assert len(st.backward_id_referenced_by[se_b.uuid]) == 0
    assert len(st.backward_id_referenced_by[se_c.uuid]) == 1
    se_a in st.backward_id_referenced_by[se_c.uuid]

    st._merge_into(se_a, se_b)

    # CHECK REFERENCE MAP (after merge):
    # c is now referenced by b
    assert se_a.uuid not in st.forward_references
    assert len(st.forward_references[se_b.uuid]) == 1
    se_c in st.forward_references[se_b.uuid]
    assert len(st.forward_references[se_c.uuid]) == 0
    assert se_a.uuid not in st.backward_references
    assert len(st.backward_references[se_b.uuid]) == 0
    assert len(st.backward_references[se_c.uuid]) == 1
    se_b in st.backward_references[se_c.uuid]

    assert se_a.uuid not in st.forward_id_references
    assert len(st.forward_id_references[se_b.uuid]) == 1
    se_c in st.forward_id_references[se_b.uuid]
    assert len(st.forward_id_references[se_c.uuid]) == 0
    assert se_a.uuid not in st.backward_id_references
    assert len(st.backward_id_references[se_b.uuid]) == 0
    assert len(st.backward_id_references[se_c.uuid]) == 1
    se_b in st.backward_id_references[se_c.uuid]

    assert se_a.uuid not in st.forward_id_referenced_by
    assert len(st.forward_id_referenced_by[se_b.uuid]) == 1
    se_c in st.forward_id_referenced_by[se_b.uuid]
    assert len(st.forward_id_referenced_by[se_c.uuid]) == 0
    assert se_a.uuid not in st.backward_id_referenced_by
    assert len(st.backward_id_referenced_by[se_b.uuid]) == 0
    assert len(st.backward_id_referenced_by[se_c.uuid]) == 1
    se_b in st.backward_id_referenced_by[se_c.uuid]

    # simple case: a -> c <- b (a & b reference c; a & b have the same target record)
    c = db.Record(name='c').add_parent("RT2")
    a = db.Record().add_parent("RT1").add_property('RT2', c)
    b = db.Record().add_parent("RT1").add_property('RT2', c)

    st = SemanticTarget([a, b], ident_adapter)
    se_a = st.se_lookup[id(a)]
    se_b = st.se_lookup[id(b)]
    se_c = st.se_lookup[id(c)]

    # CHECK REFERENCE MAP:
    # c is referenced by a & b
    assert len(st.forward_references[se_a.uuid]) == 1
    se_c in st.forward_references[se_a.uuid]
    assert len(st.forward_references[se_b.uuid]) == 1
    se_c in st.forward_references[se_b.uuid]
    assert len(st.forward_references[se_c.uuid]) == 0
    assert len(st.backward_references[se_a.uuid]) == 0
    assert len(st.backward_references[se_b.uuid]) == 0
    assert len(st.backward_references[se_c.uuid]) == 2
    se_a in st.backward_references[se_c.uuid]
    se_b in st.backward_references[se_c.uuid]

    assert len(st.forward_id_references[se_a.uuid]) == 1
    se_c in st.forward_id_references[se_a.uuid]
    assert len(st.forward_id_references[se_b.uuid]) == 1
    se_c in st.forward_id_references[se_b.uuid]
    assert len(st.forward_id_references[se_c.uuid]) == 0
    assert len(st.backward_id_references[se_a.uuid]) == 0
    assert len(st.backward_id_references[se_b.uuid]) == 0
    assert len(st.backward_id_references[se_c.uuid]) == 2
    se_a in st.backward_id_references[se_c.uuid]
    se_b in st.backward_id_references[se_c.uuid]

    assert len(st.forward_id_referenced_by[se_a.uuid]) == 1
    se_c in st.forward_id_referenced_by[se_a.uuid]
    assert len(st.forward_id_referenced_by[se_b.uuid]) == 1
    se_c in st.forward_id_referenced_by[se_b.uuid]
    assert len(st.forward_id_referenced_by[se_c.uuid]) == 0
    assert len(st.backward_id_referenced_by[se_a.uuid]) == 0
    assert len(st.backward_id_referenced_by[se_b.uuid]) == 0
    assert len(st.backward_id_referenced_by[se_c.uuid]) == 2
    se_a in st.backward_id_referenced_by[se_c.uuid]
    se_b in st.backward_id_referenced_by[se_c.uuid]

    st._merge_into(se_a, se_b)

    # CHECK REFERENCE MAP (after merge):
    # c is now referenced by b
    # (same situation as above)
    assert se_a.uuid not in st.forward_references
    assert len(st.forward_references[se_b.uuid]) == 1
    se_c in st.forward_references[se_b.uuid]
    assert len(st.forward_references[se_c.uuid]) == 0
    assert se_a.uuid not in st.backward_references
    assert len(st.backward_references[se_b.uuid]) == 0
    assert len(st.backward_references[se_c.uuid]) == 1
    se_b in st.backward_references[se_c.uuid]

    assert se_a.uuid not in st.forward_id_references
    assert len(st.forward_id_references[se_b.uuid]) == 1
    se_c in st.forward_id_references[se_b.uuid]
    assert len(st.forward_id_references[se_c.uuid]) == 0
    assert se_a.uuid not in st.backward_id_references
    assert len(st.backward_id_references[se_b.uuid]) == 0
    assert len(st.backward_id_references[se_c.uuid]) == 1
    se_b in st.backward_id_references[se_c.uuid]

    assert se_a.uuid not in st.forward_id_referenced_by
    assert len(st.forward_id_referenced_by[se_b.uuid]) == 1
    se_c in st.forward_id_referenced_by[se_b.uuid]
    assert len(st.forward_id_referenced_by[se_c.uuid]) == 0
    assert se_a.uuid not in st.backward_id_referenced_by
    assert len(st.backward_id_referenced_by[se_b.uuid]) == 0
    assert len(st.backward_id_referenced_by[se_c.uuid]) == 1
    se_b in st.backward_id_referenced_by[se_c.uuid]


def test_backward_id_referenced_by():
    # We use the reference as identifying reference in both directions. Thus the map is the same
    # for all three categories: references, id_references and id_referenced_by
    ident_a = db.RecordType().add_parent("BR").add_property("name")
    ident_b = db.RecordType().add_parent("C").add_property("is_referenced_by", ["BR"])
    ident_adapter = CaosDBIdentifiableAdapter()
    ident_adapter.register_identifiable("BR", ident_a)
    ident_adapter.register_identifiable("C", ident_b)

    referenced = db.Record(name="B").add_parent("C")
    entlist = [referenced, db.Record(name="A").add_parent("BR").add_property("ref", referenced), ]

    st = SemanticTarget(entlist, ident_adapter)
    assert st.se[1] in st.backward_id_referenced_by[st.se[0].uuid]


@patch("caoscrawler.semantic_target.cached_get_entity_by",
       new=Mock(side_effect=mock_get_entity_by))
def test_merging():
    # identifying information can be given at various locations in the hierachical tree
    # test whether an object is correctly combined for all cases
    ident_adapter = CaosDBIdentifiableAdapter()
    ident_a = db.RecordType().add_parent("A").add_property("name").add_property("a")
    ident_adapter.register_identifiable("A", ident_a)
    ident_adapter.retrieve_identified_record_for_identifiable = Mock(
        side_effect=partial(
            basic_retrieve_by_name_mock_up, known={"A": db.Record(id=1111, name="A")}))

    # merging based on id
    entlist = [
        db.Record(id=101).add_parent("A"),
        db.Record(id=101).add_parent("A")]
    st = SemanticTarget(entlist, ident_adapter)
    assert len(st.se) == 1
    assert len(st.unchecked) == 0
    assert entlist[0] in st.se[0].fragments
    assert entlist[1] in st.se[0].fragments

    # merging based on path
    entlist = [
        db.File(path='101').add_parent("A"),
        db.File(path='101').add_parent("A")]
    st = SemanticTarget(entlist, ident_adapter)
    assert len(st.se) == 1
    assert len(st.unchecked) == 0
    assert entlist[0] in st.se[0].fragments
    assert entlist[1] in st.se[0].fragments

    # merging based on identifiable
    entlist = [
        db.File(name='101').add_parent("A").add_property('a', value=1),
        db.File(name='101').add_parent("A").add_property('a', value=1)]
    st = SemanticTarget(entlist, ident_adapter)
    assert len(st.unchecked) == 2
    st.make_identifiable(st.se[0])
    st.check_remote_server(st.se[0])
    st.set_missing(st.se[0])
    assert len(st.unchecked) == 1
    st.make_identifiable(st.se[1])
    assert st.se[1].id is None
    assert st.merge_with_equivalent(st.se[1])
    assert len(st.se) == 1
    assert len(st.unchecked) == 0
    assert entlist[0] in st.se[0].fragments
    assert entlist[1] in st.se[0].fragments

    # Merging a mix. One Record needs the identifiable to be merged. But the identifying
    # information is scattered in the other case.
    entlist = [
        db.Record(id=101).add_parent("A"),
        db.Record(id=101, name='a').add_parent("A"),
        db.Record(id=101).add_parent("A").add_property('a', value=1),
        db.Record(name='a').add_parent("A").add_property('a', value=1)]

    st = SemanticTarget(entlist, ident_adapter)
    assert len(st.se) == 2
    assert len(st.unchecked) == 1
    st.make_identifiable(st.se[1])
    assert st.merge_with_equivalent(st.se[1])
    assert len(st.se) == 1
    assert len(st.unchecked) == 0
    for ii in range(4):
        assert entlist[ii] in st.se[0].fragments
